﻿using System;
using System.ComponentModel.Design;
using System.Globalization;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Task = System.Threading.Tasks.Task;

using EnvDTE;
using System.Collections.Generic;
using Microsoft.VisualStudio.TextManager.Interop;
using System.Runtime.InteropServices;
using System.Linq;

namespace VsFileMacros
{
    /// <summary>
    /// Command handler
    /// </summary>
    internal sealed class ExpandFuncProto
    {
        /// <summary>
        /// Command ID.
        /// </summary>
        public const int CommandId = 4130;

        /// <summary>
        /// Command menu group (command set GUID).
        /// </summary>
        public static readonly Guid CommandSet = new Guid("1a2cdd6e-8ed8-4a84-9dc4-c61aed04b210");

        /// <summary>
        /// VS Package that provides this command, not null.
        /// </summary>
        private readonly AsyncPackage package;

        /// <summary>
        /// Initializes a new instance of the <see cref="ExpandFuncProto"/> class.
        /// Adds our command handlers for menu (commands must exist in the command table file)
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        /// <param name="commandService">Command service to add command to, not null.</param>
        private ExpandFuncProto(AsyncPackage package, OleMenuCommandService commandService)
        {
            this.package = package ?? throw new ArgumentNullException(nameof(package));
            commandService = commandService ?? throw new ArgumentNullException(nameof(commandService));

            var menuCommandID = new CommandID(CommandSet, CommandId);
            var menuItem = new MenuCommand(this.Execute, menuCommandID);
            commandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Gets the instance of the command.
        /// </summary>
        public static ExpandFuncProto Instance
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the service provider from the owner package.
        /// </summary>
        private Microsoft.VisualStudio.Shell.IAsyncServiceProvider ServiceProvider
        {
            get
            {
                return this.package;
            }
        }

        /// <summary>
        /// Initializes the singleton instance of the command.
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        public static async Task InitializeAsync(AsyncPackage package)
        {
            // Switch to the main thread - the call to AddCommand in ExpandFuncProto's constructor requires
            // the UI thread.
            await ThreadHelper.JoinableTaskFactory.SwitchToMainThreadAsync(package.DisposalToken);

            OleMenuCommandService commandService = await package.GetServiceAsync(typeof(IMenuCommandService)) as OleMenuCommandService;
            Instance = new ExpandFuncProto(package, commandService);
        }

        /// <summary>
        /// This function is the callback used to execute the command when the menu item is clicked.
        /// See the constructor to see how the menu item is associated with this function using
        /// OleMenuCommandService service and MenuCommand class.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event args.</param>
        private void Execute(object sender, EventArgs e)
        {
            ThreadHelper.ThrowIfNotOnUIThread();
            DTE dte = Package.GetGlobalService(typeof(SDTE)) as DTE;
            string filename = dte.ActiveDocument.Name;
            int pos_ext = filename.IndexOf('.');
            string classname = (pos_ext < 0) ? filename : filename.Substring (0, pos_ext);

            var service = Package.GetGlobalService(typeof(SVsTextManager));
            var text_manager = service as IVsTextManager2;
            IVsTextView view;
            int result = text_manager.GetActiveView2(1, null, (uint)_VIEWFRAMETYPE.vftCodeWindow, out view);

            // Selection position
            Helper.get_selection(ref view, out int row_beg, out int col_beg, out int row_end, out int col_end);

            // Extends the selection to whole lines
            col_beg = 0;
            if (col_end > 0)
            {
                ++row_end;
            }
            col_end = 0;

            // Gets the selected text
            view.GetBuffer(out IVsTextLines lines);
            lines.GetLineText(row_beg, 0, row_end, 0, out string txt);
            txt = txt.Trim();

            // Handles each prototype, erase full function definitions (not
            // declarations)
            string txt_out = "";
            while (txt.Length > 0)
            {
                string proto = Helper.isolate_prototype(txt, out bool decl_flag, out int pos_end); 
                if (! decl_flag)
                {
                    pos_end = txt.IndexOf('{', pos_end);
                    pos_end = Helper.find_matching_symbol(txt, pos_end);
                }
                else
                {
                    pos_end = txt.IndexOf(';', pos_end) + 1;
                    txt_out += process_prototype(proto, classname);
                }
                txt = txt.Substring(pos_end).Trim();
            }

            // Replaces prototypes with the full functions
            lines.ReplaceLines(
               row_beg, col_beg, row_end, col_end,
               Marshal.StringToHGlobalUni(txt_out),
               txt_out.Length, new TextSpan[1]
           );
        }

        public static string process_prototype (string proto, string classname)
        {
            string txt_out = "";

            // Analyzes the basic prototype structure
            List<Helper.TopLevelPar> par_list = Helper.locate_parenthesis(proto);
            if (!Helper.locate_func_name_and_arg(par_list, out Helper.TopLevelPar par_main))
            {
                // Cannot find the function name
                return "";
            }
            bool dtor_flag = par_main.check_and_fix_dtor(proto);
            par_main.check_and_fix_operator(proto);

            // Removes some keywords from the return type
            int pos_ret_end = Helper.check_func_scope(proto, par_main, out _);
            string proto_beg = proto.Substring(0, pos_ret_end);

            proto_beg = Helper.remove_all_occurences(proto_beg, "static", _ws_list);
            proto_beg = Helper.remove_all_occurences(proto_beg, "inline", _ws_list);
            proto_beg = Helper.remove_all_occurences(proto_beg, "virtual", _ws_list);
            proto_beg = Helper.remove_all_occurences(proto_beg, "explicit", _ws_list);

            proto_beg = Helper.simplify_whitechars(proto_beg).Trim();

            txt_out += proto_beg + ((proto_beg.Length > 0) ? "\t" : "");

            // The function name
            if (classname != "")
            {
                txt_out += classname + "::";
            }
            txt_out += par_main._identifier;
            txt_out += " (";

            // Identifies the function parameters
            List<string> argraw_list = Helper.split_fnc_parameters(proto.Substring(
                par_main._pos_beg + 1,
                par_main._pos_end - par_main._pos_beg - 2
            ));

            List<Helper.ParamInfo> param_info_list = new List<Helper.ParamInfo> ();

            bool next_flag = false;
            foreach (string arg_orig in argraw_list)
            {
                string arg = arg_orig;

                if (next_flag)
                {
                    txt_out += ", ";
                }

                Helper.ParamInfo info = Helper.analyse_param(arg);
                param_info_list.Add(info);

                // Removes the default value
                if (info._pos_eq > 0)
                {
                    arg = arg.Substring(0, info._pos_eq).TrimEnd ();
                }

                txt_out += arg;

                next_flag = true;
            }

            txt_out += ")";

            // Post function parameters
            string post = proto.Substring(par_main._pos_end);
            post = Helper.remove_all_occurences(post, "final", _ws_list);
            post = Helper.remove_all_occurences(post, "override", _ws_list);
            post = Helper.simplify_whitechars(post).Trim ();
            if (post.Length > 0)
            {
                txt_out += " " + post;
            }
            txt_out += "\n";

            // If this is a constructor, add member initialisations
            Helper.check_return(proto_beg, out bool ret_flag, out bool ctor_flag);
            if (ctor_flag)
            {
                next_flag = false;
                foreach (Helper.ParamInfo info in param_info_list)
                {
                    txt_out += (next_flag ? "," : ":");
                    txt_out += "\t_" + info._name + " (" + info._name + ")\n";
                    next_flag = true;
                }

            }

            txt_out += "{\n";

            // Default body content
            // Part 1: parameter check
            string body1 = "";
            foreach (Helper.ParamInfo info in param_info_list)
            {
                body1 += "\tassert (" + info._name + " ";
                if (info._ptr_flag)
                {
                    body1 += "!= nullptr";
                }
                else
                {
                    body1 += "/*** To do ***/";
                }
                body1 += ");\n";
            }

            // Part 2: trivial implementations
            bool ret_needed_flag = ret_flag;
            string body2 = "";
            if (Helper.is_interface (classname) && ! (dtor_flag || ctor_flag))
            {
                body2 +=
                      "\t" + (ret_flag ? "return " : "") + "do_"
                    + par_main._identifier + " ("
                    + Helper.build_arg_list (param_info_list)
                    + ")\n";
                ret_needed_flag = false;
            }
            else if (par_main._identifier.IndexOf ("set_") == 0 && param_info_list.Count > 0)
            {
                int pos = par_main._identifier.IndexOf("_") + 1;
                string name = par_main._identifier.Substring(pos);
                string member = "_" + name;
                body2 += "\t" + member + " = " + param_info_list.Last()._name + ";\n";
            }
            else
            {
                bool get_flag = (par_main._identifier.IndexOf("get_") == 0);
                bool use_flag = (par_main._identifier.IndexOf("use_") == 0);
                bool has_flag = (par_main._identifier.IndexOf("has_") == 0);
                bool is_flag  = (par_main._identifier.IndexOf("is_" ) == 0);
                if (get_flag || use_flag || has_flag || is_flag)
                {
                    int pos = par_main._identifier.IndexOf("_") + 1;
                    string member = "_" + par_main._identifier.Substring(pos);
                    if (has_flag || is_flag)
                    {
                        member += "_flag";
                    }
                    body2 += "\treturn " + member + ";\n";
                    ret_needed_flag = false;
                }

                else
                {
                    body2 += "\n\t/*** To do ***/\n\n";
                }
            }

            // Return value, if not handled previously
            string body3 = "";
            if (ret_needed_flag)
            {
                body3 = "\treturn /*** To do ***/;\n";
            }

            // Merges all body parts
            string body = body1;
            body += ((body.Length > 0 && body2.Length > 0) ? "\n" : "") + body2;
            body += ((body.Length > 0 && body3.Length > 0) ? "\n" : "") + body3;

            txt_out += body + "}\n" + "\n\n\n";

            return txt_out;
        }

        static string[] keywords_ret =
        {
            "__forceinline", "inline"
        };

        static char[] _ws_list =
        {
            ' ', '\n', '\t', '\a', '\b', '\r', '\v', '\f'
        };
    }
}
