﻿// https://stackoverflow.com/questions/59854970/get-the-project-of-the-current-editing-file
// https://docs.microsoft.com/en-us/visualstudio/extensibility/inside-the-editor?view=vs-2019
// https://michaelscodingspot.com/visual-studio-2017-extension-development-tutorial-part-3-add-context-menu-get-selected-code/

using System;
using System.ComponentModel.Design;
using System.Globalization;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Task = System.Threading.Tasks.Task;

using EnvDTE;
using Microsoft.VisualStudio.TextTemplating;
using Microsoft.VisualStudio.TextTemplating.VSHost;
using Microsoft.VisualStudio.TextManager.Interop;
using System.Runtime.InteropServices;

using System.IO;
using System.Windows.Forms;
using System.Drawing;

using System.Collections.Generic;
using System.Linq;

namespace VsFileMacros
{
    /// <summary>
    /// Command handler
    /// </summary>
    internal sealed class FileTemplate
    {
        enum FileType
        {
            header,
            codeheader,
            code,

            unknown = -1
        };

        const string _license =
              "This program is free software. It comes without any warranty, to\n"
            + "the extent permitted by applicable law.You can redistribute it\n"
            + "and/or modify it under the terms of the Do What The Fuck You Want\n"
            + "To Public License, Version 2, as published by Sam Hocevar. See\n"
            + "http://www.wtfpl.net/ for more details.\n";

        const string _author = "Laurent de Soras";

        const int _tab_len = 3;
        const int _indent_len = 15;

        /// <summary>
        /// Command ID.
        /// </summary>
        public const int CommandId = 0x0100;

        /// <summary>
        /// Command menu group (command set GUID).
        /// </summary>
        public static readonly Guid CommandSet = new Guid("1a2cdd6e-8ed8-4a84-9dc4-c61aed04b210");

        /// <summary>
        /// VS Package that provides this command, not null.
        /// </summary>
        private readonly AsyncPackage package;

        /// <summary>
        /// Initializes a new instance of the <see cref="FileTemplate"/> class.
        /// Adds our command handlers for menu (commands must exist in the command table file)
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        /// <param name="commandService">Command service to add command to, not null.</param>
        private FileTemplate(AsyncPackage package, OleMenuCommandService commandService)
        {
            this.package = package ?? throw new ArgumentNullException(nameof(package));
            commandService = commandService ?? throw new ArgumentNullException(nameof(commandService));

            var menuCommandID = new CommandID(CommandSet, CommandId);
            var menuItem = new MenuCommand(this.Execute, menuCommandID);
            commandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Gets the instance of the command.
        /// </summary>
        public static FileTemplate Instance
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the service provider from the owner package.
        /// </summary>
        private Microsoft.VisualStudio.Shell.IAsyncServiceProvider ServiceProvider
        {
            get
            {
                return this.package;
            }
        }

        /// <summary>
        /// Initializes the singleton instance of the command.
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        public static async Task InitializeAsync(AsyncPackage package)
        {
            // Switch to the main thread - the call to AddCommand in Command1's constructor requires
            // the UI thread.
            await ThreadHelper.JoinableTaskFactory.SwitchToMainThreadAsync(package.DisposalToken);

            OleMenuCommandService commandService = await package.GetServiceAsync(typeof(IMenuCommandService)) as OleMenuCommandService;
            Instance = new FileTemplate(package, commandService);
        }

        /// <summary>
        /// This function is the callback used to execute the command when the menu item is clicked.
        /// See the constructor to see how the menu item is associated with this function using
        /// OleMenuCommandService service and MenuCommand class.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event args.</param>
        private void Execute(object sender, EventArgs e)
        {
            ThreadHelper.ThrowIfNotOnUIThread();
            DTE dte = Package.GetGlobalService(typeof(SDTE)) as DTE;
            string filename = dte.ActiveDocument.Name;

            // Finds the file type (.h, .hpp or .cpp)
            FileType file_type = get_file_type(filename);
            if (file_type == FileType.unknown)
            {
                MessageBox.Show("File extension not supported.", "Error");
                return;
            }

            // If the file is a header (.h), checks if there is a companion
            // code header file (.hpp)
            bool hpp_flag = false;
            if (file_type == FileType.header)
            {
                hpp_flag = File.Exists(dte.ActiveDocument.FullName + "pp");
            }

            // Finds the class name
            string classname = get_classname(filename);
            if (classname == "")
            {
                MessageBox.Show("Cannot extract a class name from the filename.", "Error");
                return;
            }

            // Ask user for a namespace
            List<string> nspace_list = prompt_namespaces();

            // Gets the current year
            string year = $"{DateTime.Today.Year}";

            // Build the right template depending on the type
            string ins_str = "";
            switch (file_type)
            {
                case FileType.header:
                    ins_str = build_tpl_h(filename, classname, nspace_list, year, hpp_flag);
                    break;
                case FileType.codeheader:
                    ins_str = build_tpl_hpp(filename, classname, nspace_list, year);
                    break;
                case FileType.code:
                    ins_str = build_tpl_cpp(filename, classname, nspace_list, year);
                    break;
                default:
                    throw new Exception("Unexpected FileType value");
            }

            // Inserts it at the cursor location or replaces the selection
            var service = Package.GetGlobalService(typeof(SVsTextManager));
            var text_manager = service as IVsTextManager2;
            IVsTextView view;
            int result = text_manager.GetActiveView2(1, null, (uint)_VIEWFRAMETYPE.vftCodeWindow, out view);
            Helper.get_selection (ref view, out int row_beg, out int col_beg, out int row_end, out int col_end);

            IVsTextLines lines;
            view.GetBuffer(out lines);
            lines.ReplaceLines(
                row_beg, col_beg, row_end, col_end,
                Marshal.StringToHGlobalUni(ins_str),
                ins_str.Length, new TextSpan[1]
            );
        }

        // https://www.csharp-examples.net/inputbox/
        private static DialogResult InputBox(string title, string promptText, ref string value)
        {
            Form form = new Form();
            Label label = new Label();
            TextBox textBox = new TextBox();
            Button buttonOk = new Button();
            Button buttonCancel = new Button();

            form.Text = title;
            label.Text = promptText;
            textBox.Text = value;

            buttonOk.Text = "OK";
            buttonCancel.Text = "Cancel";
            buttonOk.DialogResult = DialogResult.OK;
            buttonCancel.DialogResult = DialogResult.Cancel;

            label.SetBounds(9, 20, 372, 13);
            textBox.SetBounds(12, 36, 372, 20);
            buttonOk.SetBounds(228, 72, 75, 23);
            buttonCancel.SetBounds(309, 72, 75, 23);

            label.AutoSize = true;
            textBox.Anchor = textBox.Anchor | AnchorStyles.Right;
            buttonOk.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;
            buttonCancel.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;

            form.ClientSize = new Size(396, 107);
            form.Controls.AddRange(new Control[] { label, textBox, buttonOk, buttonCancel });
            form.ClientSize = new Size(Math.Max(300, label.Right + 10), form.ClientSize.Height);
            form.FormBorderStyle = FormBorderStyle.FixedDialog;
            form.StartPosition = FormStartPosition.CenterScreen;
            form.MinimizeBox = false;
            form.MaximizeBox = false;
            form.AcceptButton = buttonOk;
            form.CancelButton = buttonCancel;

            DialogResult dialogResult = form.ShowDialog();
            value = textBox.Text;
            return dialogResult;
        }

        private static List<string> prompt_namespaces()
        {
            string nspace_str = "";
            InputBox("Namespaces", "Enter namespaces sparated with space, slash, backslash or ::", ref nspace_str);

            string[] delim = { " ", "/", "\\", "::", "\t" };
            string[] berk = nspace_str.Split(delim, System.StringSplitOptions.RemoveEmptyEntries);

            return new List<string>(berk);
        }

        private static FileType get_file_type(string filename)
        {
            string[] nameandext = filename.Split('.');
            if (nameandext.Length < 2)
            {
                return FileType.unknown;
            }
            string fileext = nameandext[nameandext.Length - 1].ToLowerInvariant();
            return
                  (fileext == "h") ? FileType.header
                : (fileext == "hpp") ? FileType.codeheader
                : (fileext == "cpp") ? FileType.code
                : FileType.unknown;
        }

        private static string get_classname(string filename)
        {
            int end_pos = filename.LastIndexOf('.');
            if (end_pos < 0)
            {
                return "";
            }
            return filename.Substring(0, end_pos);
        }

        private static string build_tpl_h(string filename, string classname, List<string> nspace_list, string year, bool hpp_flag)
        {
            string guard = make_guard_prefix(classname, nspace_list) + "HEADER_INCLUDED";
            string str =
                  make_title_block (filename, year)
                + "#pragma once\n"
                + "#if ! defined (" + guard + ")\n"
                + "#define " + guard + "\n"
                + "\n\n\n"
                + make_separator("INCLUDE FILES")
                + "\n\n\n"
                + make_namespaces_open(nspace_list)
                + "class " + classname + "\n"
                + "{\n"
                + "\n"
                + make_separator("PUBLIC")
                + "\n"
                + "public:\n"
                + "\n\n\n"
                + make_separator("PROTECTED")
                + "\n"
                + "protected:\n"
                + "\n\n\n"
                + make_separator("PRIVATE")
                + "\n"
                + "private:\n"
                + "\n\n\n"
                + make_separator("FORBIDDEN MEMBER FUNCTIONS")
                + "\n"
                + "private:\n"
                + "\n"
                + make_forbidden_fnc(classname)
                + "\n"
                + "}; // class " + classname + "\n"
                + "\n\n\n"
                + make_namespaces_close(nspace_list)
                + (hpp_flag ? "" : "//")
                + "#include \"" + make_path_prefix(classname, nspace_list) + ".hpp\"\n"
                + "\n\n\n"
                + "#endif // " + guard + "\n"
                + "\n\n\n"
                + make_separator("EOF");

            return str;
        }

        private static string build_tpl_hpp(string filename, string classname, List<string> nspace_list, string year)
        {
            string guard = make_guard_prefix(classname, nspace_list) + "CODEHEADER_INCLUDED";
            string str =
                  make_title_block(filename, year)
                + "#pragma once\n"
                + "#if ! defined (" + guard + ")\n"
                + "#define " + guard + "\n"
                + "\n\n\n"
                + make_separator("INCLUDE FILES")
                + "\n\n\n"
                + make_code_body(nspace_list)
                + "#endif // " + guard + "\n"
                + "\n\n\n"
                + make_separator("EOF");
            return str;
        }

        private static string build_tpl_cpp(string filename, string classname, List<string> nspace_list, string year)
        {
            string str =
                  make_title_block(filename, year)
                + make_separator("INCLUDE FILES")
                + "\n"
                + "#include \"" + make_path_prefix(classname, nspace_list) + ".h\"\n"
                + "\n\n\n"
                + make_code_body (nspace_list)
                + make_separator("EOF");
            return str;
        }

        private static string make_title_block (string filename, string year)
        {
            string str =
                   "/*****************************************************************************\n"
                 + "\n"
                 + "        " + filename + "\n"
                 + "        Author: " + _author + ", " + year + "\n"
                 + "\n"
                 + "--- Legal stuff ---\n"
                 + "\n"
                 + _license
                 + "\n"
                 + $"*Tab={_tab_len}***********************************************************************/\n"
                 + "\n\n\n";

            return str;
        }
        private static string make_guard_prefix(string classname, List<string> nspace_list)
        {
            return make_string_prefix(classname, nspace_list, "_") + "_";
        }

        private static string make_path_prefix(string classname, List<string> nspace_list)
        {
            return make_string_prefix (classname, nspace_list, "/");
        }

        private static string make_string_prefix(string classname, List<string> nspace_list, string delim)
        {
            string prefix = "";
            foreach (string s in nspace_list)
            {
                prefix += s + delim;
            }
            prefix += classname;
            return prefix;
        }

        private static string make_code_body (List<string> nspace_list)
        {
            return
                  make_namespaces_open(nspace_list)
                + make_separator("PUBLIC")
                + "\n\n\n"
                + make_separator("PROTECTED")
                + "\n\n\n"
                + make_separator("PRIVATE")
                + "\n\n\n"
                + make_namespaces_close(nspace_list);
        }

        private static string make_separator(string content)
        {
            string sep = "/*\\\\\\ " + content + " ";
            sep += new string('\\', 76 - sep.Length);
            sep += "*/\n";
            return sep;
        }

        private static string make_namespaces_open(List<string> nspace_list)
        {
            string content = "";
            foreach (string s in nspace_list)
            {
                content += "namespace " + s + "\n{\n";
            }
            if (nspace_list.Count > 0)
            {
                content += "\n\n\n";
            }
            return content;
        }

        private static string make_namespaces_close(List<string> nspace_list)
        {
            string content = "";
            foreach (string s in Enumerable.Reverse(nspace_list))
            {
                content += "}  // namespace " + s + "\n";
            }
            if (nspace_list.Count > 0)
            {
                content += "\n\n\n";
            }
            return content;
        }

        private static string make_forbidden_fnc (string classname)
        {
            int cnl = classname.Length;
            string ret_n = align_member("");
            string ret_c = align_member (classname + " &");
            string ret_b = align_member("bool");
            string content = "";
            content += ret_n + term_member_del(classname + " ()", cnl);
            content += ret_n + term_member_del(classname + " (const " + classname + " &other)", cnl);
            content += ret_n + term_member_del(classname + " (" + classname + " &&other)", cnl);
            content += ret_c + term_member_del("operator = (const " + classname + " &other)", cnl);
            content += ret_c + term_member_del("operator = (" + classname + " &&other)", cnl);
            content += ret_b + term_member_del("operator == (const " + classname + " &other) const", cnl);
            content += ret_b + term_member_del("operator != (const " + classname + " &other) const", cnl);
            return content;
        }

        private static string align_member (string ret_val)
        {
            string content = "\t" + ret_val;
            if (ret_val.Length >= _indent_len)
            {
                content += "\n" + align_member("");
            }
            else
            {
                content += new string(' ', _indent_len - ret_val.Length);
            }
            return content;
        }

        private static string term_member_del (string member, int classname_len)
        {
            int align = 33 + classname_len;
            align = (align > 50) ? 50 : align;
            string content = member;
            int fill_len = align - member.Length;
            fill_len = (fill_len < 0) ? 0 : fill_len;
            content += new string (' ', fill_len) + " = delete;\n";
            return content;
        }
    }
}



/*
            // Show a message box to prove we were here
            VsShellUtilities.ShowMessageBox(
                this.package,
                message,
                title,
                OLEMSGICON.OLEMSGICON_INFO,
                OLEMSGBUTTON.OLEMSGBUTTON_OK,
                OLEMSGDEFBUTTON.OLEMSGDEFBUTTON_FIRST
            );
*/

