﻿using System;
using System.ComponentModel.Design;
using System.Globalization;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Task = System.Threading.Tasks.Task;

using System.Collections.Generic;
using Microsoft.VisualStudio.TextManager.Interop;
using System.Runtime.InteropServices;

namespace VsFileMacros
{
    /// <summary>
    /// Command handler
    /// </summary>
    internal sealed class FuncDesc
    {
        /// <summary>
        /// Command ID.
        /// </summary>
        public const int CommandId = 4129;

        /// <summary>
        /// Command menu group (command set GUID).
        /// </summary>
        public static readonly Guid CommandSet = new Guid("1a2cdd6e-8ed8-4a84-9dc4-c61aed04b210");

        /// <summary>
        /// VS Package that provides this command, not null.
        /// </summary>
        private readonly AsyncPackage package;

        /// <summary>
        /// Initializes a new instance of the <see cref="FuncDesc"/> class.
        /// Adds our command handlers for menu (commands must exist in the command table file)
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        /// <param name="commandService">Command service to add command to, not null.</param>
        private FuncDesc(AsyncPackage package, OleMenuCommandService commandService)
        {
            this.package = package ?? throw new ArgumentNullException(nameof(package));
            commandService = commandService ?? throw new ArgumentNullException(nameof(commandService));

            var menuCommandID = new CommandID(CommandSet, CommandId);
            var menuItem = new MenuCommand(this.Execute, menuCommandID);
            commandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Gets the instance of the command.
        /// </summary>
        public static FuncDesc Instance
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the service provider from the owner package.
        /// </summary>
        private Microsoft.VisualStudio.Shell.IAsyncServiceProvider ServiceProvider
        {
            get
            {
                return this.package;
            }
        }

        /// <summary>
        /// Initializes the singleton instance of the command.
        /// </summary>
        /// <param name="package">Owner package, not null.</param>
        public static async Task InitializeAsync(AsyncPackage package)
        {
            // Switch to the main thread - the call to AddCommand in Command1's constructor requires
            // the UI thread.
            await ThreadHelper.JoinableTaskFactory.SwitchToMainThreadAsync(package.DisposalToken);

            OleMenuCommandService commandService = await package.GetServiceAsync(typeof(IMenuCommandService)) as OleMenuCommandService;
            Instance = new FuncDesc(package, commandService);
        }

        /// <summary>
        /// This function is the callback used to execute the command when the menu item is clicked.
        /// See the constructor to see how the menu item is associated with this function using
        /// OleMenuCommandService service and MenuCommand class.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event args.</param>
        private void Execute(object sender, EventArgs e)
        {
            ThreadHelper.ThrowIfNotOnUIThread();

            var service = Package.GetGlobalService(typeof(SVsTextManager));
            var text_manager = service as IVsTextManager2;
            IVsTextView view;
            int result = text_manager.GetActiveView2(1, null, (uint)_VIEWFRAMETYPE.vftCodeWindow, out view);

            // Gets the cursor/selection position
            Helper.get_selection(ref view, out int row_beg, out int col_beg, out int row_end, out int col_end);

            // Gets the text at the cursor (up to 100 lines)
            view.GetBuffer(out IVsTextLines lines);
            lines.GetLineCount(out int nbr_lines);
            int row_stop = row_beg + 100;
            row_stop = (row_stop >= nbr_lines) ? nbr_lines - 1 : row_stop;
            lines.GetLineText (row_beg, 0, row_stop, 0, out string line);

            // Finds the end of the prototype
            line = Helper.isolate_prototype(line, out _, out _);

            Helper.parse_prototype(
                line, out string func_name,
                out List<string> arg_list, out List<bool> in_flag_list,
                out bool ret_flag, out Helper.Type type, out bool noexcept_flag
            );

            string content =
                  "/*\n"
                + "==============================================================================\n"
                + "Name: ";
            if (type != Helper.Type.normal)
            {
                content += ((type == Helper.Type.ctor) ? "c" : "d") + "tor\n";
            }
            else
            {
                content += func_name + "\n";
                content += "Description:\n";
            }
            if (in_flag_list.Contains (true))
            {
                content += "Input parameters:\n";
                content += print_parameters(arg_list, in_flag_list, true);
            }
            if (in_flag_list.Contains (false))
            {
                content += "Input/output parameters:\n";
                content += print_parameters(arg_list, in_flag_list, false);
            }
            if (ret_flag)
            {
                content += "Returns:\n";
            }
            if (type != Helper.Type.dtor)
            {
                content += "Throws:" + (noexcept_flag ? " Nothing" : "") + "\n";
            }
            content +=
                  "==============================================================================\n"
                + "*/\n";

            lines.ReplaceLines(
                row_beg, 0, row_beg, 0,
                Marshal.StringToHGlobalUni(content),
                content.Length, new TextSpan[1]
            );
        }

        static private string print_parameters (in List<string> arg_list, in List<bool> in_flag_list, bool in_flag)
        {
            string txt = "";

            for (int idx = 0; idx < arg_list.Count; ++idx)
            {
                if (in_flag_list[idx] == in_flag)
                {
                    txt += "- " + arg_list[idx] + ":\n";
                }
            }

            return txt;
        }
    }
}
