﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.VisualStudio.TextManager.Interop;

namespace VsFileMacros
{
    class Helper
    {
        public static void get_selection (ref IVsTextView view, out int row_beg, out int col_beg, out int row_end, out int col_end)
        {
            view.GetSelection(out row_beg, out col_beg, out row_end, out col_end);
            // end could be before beginning
            if (row_end < row_beg || (row_end == row_beg && col_end < col_beg))
            {
                var r_tmp = row_end;
                var c_tmp = col_end;
                row_end = row_beg;
                col_end = col_beg;
                row_beg = r_tmp;
                col_beg = c_tmp;
            }
        }

        /*
On ne passe que le prototype de la fonction.
L'appelant peut detecter la fin du proto par le premier caractere '{' ou ';'
trouve. Peut-etre peut-on sauter tous les trucs qui s'ouvrent et se ferment :
(), [], {}, ainsi que <> ?

La difficulte est de sauter tous les decorateurs, a tous les endroits
possibles.

De ce fait, trouver le nom de la fonction est delicat.

Pieges devant :
__attribute__()
__declspec()
decltype()
template<>

Pieges apres :
throw()
noexcept()
=

Pieges de chaque cote :
[[]]

        */

        public enum Type
        {
            normal = 0,
            ctor,
            dtor
        };

        // Finds the end of the prototype:
        // - '{' or ';'
        // - ':' after a parenthesis pair, but not "::".
        // Skips any opening-closing character pairs inbetween.
        // The returned string includes the closing parenthesis, nothing after.
        // pos_end contains the position just after the closing parenthesis.
        public static string isolate_prototype (string txt, out bool decl_flag, out int pos_end)
        {
            decl_flag = false;
            int pos = 0;
            bool par_flag = false;
            while (pos < txt.Length)
            {
                char c = txt[pos];
                if (c == ':' && pos < txt.Length - 1 && txt[pos + 1] == ':')
                {
                    pos += 2;
                }
                else if (c == ';')
                {
                    decl_flag = true;
                    break;
                }
                else if (c == '{' || c == ';' || (c == ':' && par_flag))
                {
                    break;
                }
                else if (_char_pair_dict.ContainsKey(c))
                {
                    pos = find_matching_symbol(txt, pos);
                    par_flag |= (c == '(');
                }
                else
                {
                    ++pos;
                }
            }

            txt = txt.Substring(0, pos).TrimEnd();
            pos_end = txt.Length;
            return txt.TrimStart ();
        }

        public static void parse_prototype (string line, out string func_name, out List <string> arg_list, out List <bool> in_flag_list, out bool ret_flag, out Type type, out bool noexcept_flag)
        {
            func_name = "";
            arg_list = new List<string> ();
            in_flag_list = new List<bool>();
            ret_flag = false;
            type = Type.normal;
            noexcept_flag = false;

            List<TopLevelPar> par_list = locate_parenthesis(line);
            if (! locate_func_name_and_arg (par_list, out TopLevelPar par_main))
            {
                // Cannot find the function name
                return;
            }
            bool dtor_flag = par_main.check_and_fix_dtor(line);

            // Operator functions are a bit special, we have to fix what we
            // found.
            par_main.check_and_fix_operator(line);

            // We've got the function name
            func_name = par_main._identifier;
            int pos = par_main._pos_idn;

            // Destructor? We can leave immediately
            if (dtor_flag)
            {
                type = Type.dtor;
                return;
            }

            // Class name, if specified
            // Then skips other nested classnames
            int pos_ret_end = check_func_scope(line, par_main, out string classname);

            // Finds if there is a return type.
            string line_pre = line.Substring(0, pos_ret_end);
            check_return(line_pre, out ret_flag, out bool ctor_flag);
            if (ctor_flag)
            {
                type = Type.ctor;
            }

            // Now identifies the function parameters
            List<string> argraw_list = split_fnc_parameters(line.Substring (
                par_main._pos_beg + 1,
                par_main._pos_end - par_main._pos_beg - 2
            ));

            // Iterates on each parameters
            foreach (string param in argraw_list)
            {
                ParamInfo param_info = analyse_param(param);
                arg_list.Add(param_info._name);
                in_flag_list.Add(param_info._in_flag);
            }

            // Is it noexcept?
            noexcept_flag =
                line.Substring (par_main._pos_end).Contains("noexcept");
        }

        // Finds all top-levels parenthesis pairs
        // Associates each pair with the preceding identifier, if existing
        public static List<TopLevelPar> locate_parenthesis (string line)
        {
            List<TopLevelPar> par_list = new List<TopLevelPar>();
            int pos = 0;
            int pos_last = 0; // Position past the previous closing character (not necessary a parenthesis)
            while (pos < line.Length)
            {
                char c = line[pos];
                if (_char_pair_dict.ContainsKey(c))
                {
                    int pos_beg = pos;
                    pos = find_matching_symbol(line, pos);
                    if (c == '(')
                    {
                        TopLevelPar tl;
                        tl._pos_beg = pos_beg;
                        tl._pos_end = pos;
                        tl._identifier = find_last_identifier(
                            line.Substring(pos_last, pos_beg - pos_last),
                            out tl._pos_idn
                        );
                        tl._pos_idn += pos_last;
                        par_list.Add(tl);
                    }
                    pos_last = pos;
                }
                else
                {
                    ++pos;
                }
            }

            return par_list;
        }

        public static bool locate_func_name_and_arg (List<TopLevelPar> par_list, out TopLevelPar par)
        {
            par = new TopLevelPar();

            // Removes all "keyword()"-like stuff
            int pos = par_list.Count - 1;
            while (pos >= 0)
            {
                if (_keywords.Contains(par_list[pos]._identifier))
                {
                    par_list.RemoveAt(pos);
                }
                --pos;
            }

            if (par_list.Count == 0)
            {
                // Cannot find the function name
                return false;
            }

            par = par_list.Last();

            return true;
        }

        // Finds the inner scope, if explicitely specified.
        // Then skips other outer classnames and namespaces
        // Returns the position before the scopes, just after the last
        // non-whitespace character of the return section.
        public static int check_func_scope(string txt, TopLevelPar par, out string classname)
        {
            classname = "";
            int pos = skip_whitespaces(txt, par._pos_idn, -1);
            string ident = "";
            while (pos >= 2 && txt.Substring(pos - 2, 2) == "::")
            {
                pos -= 2;
                ident = find_last_identifier(txt.Substring(0, pos), out pos);
                if (classname.Length == 0)
                {
                    classname = ident;
                }
                pos = skip_whitespaces(txt, pos, -1);
            }

            return pos;
        }

        public struct TopLevelPar
        {
            public int _pos_beg; // Opening parenthesis
            public int _pos_end; // Past closing parenthesis
            public int _pos_idn; // Identifier position
            public string _identifier; // Just before

            public bool check_and_fix_dtor (string txt)
            {
                if (_pos_idn > 0 && txt [_pos_idn] == '~')
                {
                    --_pos_idn;
                    return true;
                }

                return false;
            }

            public bool check_and_fix_operator (string txt)
            {
                if (_identifier.Length == 0)
                {
                    // Goes backwards to find the end of an identifier
                    int pos = _pos_idn;
                    while (pos > 0)
                    {
                        char c = txt[pos - 1];
                        if (Char.IsLetterOrDigit(c) || c == '_')
                        {
                            string ident = find_last_identifier(
                                txt.Substring (0, pos), out pos
                            );
                            if (ident != "operator")
                            {
                                return false;
                            }
                            else
                            {
                                _identifier = txt.Substring(pos, _pos_idn - pos);
                                _identifier = simplify_whitechars(_identifier.TrimEnd());
                                _pos_idn = pos;

                                return true;
                            }
                            
                        }

                        --pos;
                    }
                }

                return false;
            }
        };

        public static string[] _keywords =
        {
            "decltype", "throw", "noexcept",
            "template", "explcit", "inline", "static", "virtual",
            "__attribute__", "__declspec"
        };

        // Input position should contain the opening symbol to match
        // Returns the position just after the matched closing symbol.
        // Skips all paired symbols.
        // If closing symbol is not found, returns the end of the line.
        public static int find_matching_symbol (string line, int pos)
        {
            char c_ref = line[pos];
            char c_end = _char_pair_dict[c_ref];
            ++pos;
            while (pos < line.Length)
            {
                char c_tst = line[pos];
                if (c_tst == c_end)
                {
                    ++pos;
                    break;
                }
                else if (_char_pair_dict.ContainsKey (c_tst))
                {
                    pos = find_matching_symbol(line, pos);
                }
                else 
                {
                    ++pos;
                }
            }

            return pos;
        }

        public static Dictionary<char, char> _char_pair_dict = new Dictionary<char, char>
        {
            {  '(', ')' },
            {  '[', ']' },
            {  '{', '}' },
            {  '<', '>' }
        };

        public static int find_matching_symbol_rev (string line, int pos)
        {
            char c_ref = line[pos - 1];
            char c_end = _char_pair_dict_rev[c_ref];
            --pos;
            while (pos > 0)
            {
                char c_tst = line[pos - 1];
                if (c_tst == c_end)
                {
                    --pos;
                    break;
                }
                else if (_char_pair_dict_rev.ContainsKey(c_tst))
                {
                    pos = find_matching_symbol_rev(line, pos);
                }
                else
                {
                    --pos;
                }
            }

            return pos;
        }

        public static Dictionary<char, char> _char_pair_dict_rev = new Dictionary<char, char>
        {
            {  ')', '(' },
            {  ']', '[' },
            {  '}', '{' },
            {  '>', '<' }
        };

        // Assumes the end of the string is an identifier.
        // Extracts it and returns its beginning position.
        public static string find_last_identifier(string line, out int ident_pos)
        {
            line = line.TrimEnd();
            int len = 0;
            while (len < line.Length)
            {
                char c = line[line.Length - 1 - len];
                if (! Char.IsLetterOrDigit (c) && c != '_')
                {
                    break;
                }
                ++len;
            }
            ident_pos = line.Length - len;
            return line.Substring(ident_pos, len);
        }

        // dir = 1 or -1
        // Forward: returns the first non-ws position
        // Backward: returns past the last non-ws position (pos is also past
        // the search position)
        // May return 0 or line.Length if the line ends with whitespaces.
        public static int skip_whitespaces (string line, int pos, int dir)
        {
            if (dir < 0)
            {
                while (pos > 0)
                {
                    if (! Char.IsWhiteSpace(line[pos - 1]))
                    {
                        break;
                    }
                    --pos;
                }
            }
            else
            {
                while (pos < line.Length)
                {
                    if (! Char.IsWhiteSpace(line[pos]))
                    {
                        break;
                    }
                    ++pos;
                }
            }

            return pos;
        }

        public static int skip_templates (string line)
        {
            const string s = "template";
            int pos = line.LastIndexOf(s);
            if (pos < 0)
            {
                pos = 0;
            }
            else
            {
                pos += s.Length;
                pos = skip_whitespaces(line, pos, +1);
                if (line [pos] == '<')
                {
                    pos = find_matching_symbol(line, pos);
                }
            }

            return pos;
        }

        // Jumps to the next identifier (starting from pos), skips any
        // character which is not part of an identifier.
        // Note: does not skip correctly numbers or litterals.
        public static int jump_to_next_identifier(string line, int pos)
        {
            while (pos < line.Length)
            {
                char c = line[pos];
                if (Char.IsLetterOrDigit(c) || c == '_')
                {
                    break;
                }
                ++pos;
            }


            return pos;
        }

        public static string extract_identifier (string line, int pos)
        {
            int pos_beg = pos;
            while (pos < line.Length)
            {
                char c = line[pos];
                if (! Char.IsLetterOrDigit(c) && c != '_')
                {
                    break;
                }
                ++pos;
            }

            return line.Substring(pos_beg, pos - pos_beg);
        }

        // Returns whitespace-trimmed parameters, no comma.
        public static List<string> split_fnc_parameters (string line)
        {
            List<string> arg_list = new List<string> ();

            int pos = 0;
            int split_pos = 0; // Past the latest comma
            while (pos < line.Length)
            {
                char c = line[pos];
                if (_char_pair_dict.ContainsKey(c))
                {
                    pos = find_matching_symbol(line, pos);
                }
                else
                {
                    ++pos;
                    if (c == ',')
                    {
                        arg_list.Add(line.Substring(split_pos, pos - 1 - split_pos).Trim());
                        split_pos = pos;
                    }
                }
            }

            if (split_pos < line.Length)
            {
                arg_list.Add(line.Substring(split_pos, line.Length - split_pos).Trim ());
            }

            return arg_list;
        }

        // If the delimiter list is empty, no delimiter is required.
        public static string remove_all_occurences (string txt, string keyword, char[] delim_list)
        {
            int scan_pos = 0;
            int pos = 0;
            while (true)
            {
                pos = txt.IndexOf(keyword, scan_pos);
                if (pos < 0)
                {
                    break;
                }
                string part_l = txt.Substring(0, pos);
                string part_r = txt.Substring(pos + keyword.Length);
                bool remove_flag = true;
                if (delim_list.Length > 0)
                {
                    if (   (part_l.Length > 0 && part_l.LastIndexOfAny(delim_list) != part_l.Length - 1)
                        || (part_r.Length > 0 && part_r.IndexOfAny(delim_list) != 0))
                    {
                        remove_flag = false;
                    }
                }
                if (remove_flag)
                {
                    txt = part_l + part_r;
                    scan_pos = pos;
                }
                else
                {
                    scan_pos = pos + 1;
                }
            }

            return txt;
        }

        public static string simplify_whitechars (string txt)
        {
            string txt_out = "";
            char prev_c = ' ';
            foreach (char c in txt)
            {
                bool ws_flag = Char.IsWhiteSpace(c);
                if (! (ws_flag && Char.IsWhiteSpace(prev_c)))
                {
                    txt_out += ws_flag ? ' ' : c;
                }
                prev_c = c;
            }
            return txt_out;
        }

        public static ParamInfo analyse_param (string param)
        {
            ParamInfo info;
            info._ptr_flag = false;

            int pos_p_end = param.Length;

            // Skips default value
            info._pos_eq = param.LastIndexOf('=');
            if (info._pos_eq > 0)
            {
                pos_p_end = skip_whitespaces(param, info._pos_eq - 1, -1);
            }

            // Skips brackets
            while (param[pos_p_end - 1] == ']')
            {
                pos_p_end = find_matching_symbol_rev(param, pos_p_end);
                pos_p_end = skip_whitespaces(param, pos_p_end, -1);
                info._ptr_flag = true;
            }

            // Identifier name
            info._name = find_last_identifier(
                param.Substring(0, pos_p_end), out int pos
            );

            // Is it const?
            string qualifiers = param.Substring(0, pos);
            info._ptr_flag |= (qualifiers.IndexOf('*') >= 0);
            info._ref_flag = qualifiers.IndexOf('&') >= 0;
            bool ref_flag = (info._ref_flag || info._ptr_flag);
            /*** To do:
            this is not very accurate. The const qualifier should apply only
            on the last * or &.
            ***/
            info._in_flag = (!ref_flag || qualifiers.Contains("const"));

            return info;
        }
        public struct ParamInfo
        {
            public int _pos_eq; // Negative if there is not default value
            public string _name;
            public bool _in_flag;
            public bool _ptr_flag;
            public bool _ref_flag;
        };

        public static bool is_interface (string classname)
        {
            return (classname.IndexOf("Interface") >= 0);
        }

        public static void check_return (string pre_fnc_name, out bool ret_flag, out bool ctor_flag)
        {
            ret_flag = false;
            ctor_flag = false;

            // Skips first any template<> at the beginning
            int pos = skip_templates(pre_fnc_name);

            // The return type is assumed being the first identifier which
            // is not a qualifier keyword.
            bool qualifier_flag = false;
            string ident = "";
            int ident_pos;
            do
            {
                pos = jump_to_next_identifier(pre_fnc_name, pos);
                ident_pos = pos;
                ident = extract_identifier(pre_fnc_name, pos);
                qualifier_flag = _keywords.Contains(ident);
                pos += ident.Length;
            }
            while (pos < pre_fnc_name.Length && qualifier_flag);

            // Checks the return value
            if (ident.Length == 0 || qualifier_flag)
            {
                ctor_flag = true;
            }
            else
            {
                ret_flag = (pre_fnc_name.Substring(ident_pos).Trim() != "void");
            }
        }

        public static string build_arg_list (List <ParamInfo> param_info_list)
        {
            string txt = "";
            bool next_flag = false;
            foreach (ParamInfo info in param_info_list)
            {
                txt += (next_flag ? ", " : "") + info._name;
                next_flag = true;
            }
            return txt;
        }

    } // class Helper
} // namespace VsFileMacros
